# strategies.py
# Copyright (C) 2005, 2006, 2007 Michael Bayer mike_mp@zzzcomputing.com
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

"""sqlalchemy.orm.interfaces.LoaderStrategy implementations, and related MapperOptions."""

from sqlalchemy import sql, schema, util, exceptions, sql_util, logging
from sqlalchemy.orm import mapper, query
from sqlalchemy.orm.interfaces import *
from sqlalchemy.orm import session as sessionlib
from sqlalchemy.orm import util as mapperutil
import random


class ColumnLoader(LoaderStrategy):
    def init(self):
        super(ColumnLoader, self).init()
        self.columns = self.parent_property.columns
        self._should_log_debug = logging.is_debug_enabled(self.logger)
        
    def setup_query(self, context, eagertable=None, parentclauses=None, **kwargs):
        for c in self.columns:
            if parentclauses is not None:
                context.statement.append_column(parentclauses.aliased_column(c))
            else:
                context.statement.append_column(c)
        
    def init_class_attribute(self):
        self.logger.info("register managed attribute %s on class %s" % (self.key, self.parent.class_.__name__))
        coltype = self.columns[0].type
        sessionlib.attribute_manager.register_attribute(self.parent.class_, self.key, uselist=False, copy_function=coltype.copy_value, compare_function=coltype.compare_values, mutable_scalars=self.columns[0].type.is_mutable())

    def process_row(self, selectcontext, instance, row, identitykey, isnew):
        if isnew:
            if self._should_log_debug:
                self.logger.debug("populating %s with %s/%s" % (mapperutil.attribute_str(instance, self.key), row.__class__.__name__, self.columns[0].key))
            instance.__dict__[self.key] = row[self.columns[0]]
        
ColumnLoader.logger = logging.class_logger(ColumnLoader)

class DeferredColumnLoader(LoaderStrategy):
    """Describes an object attribute that corresponds to a table
    column, which also will *lazy load* its value from the table.

    This is per-column lazy loading.
    """
    
    def init(self):
        super(DeferredColumnLoader, self).init()
        self.columns = self.parent_property.columns
        self.group = self.parent_property.group
        self._should_log_debug = logging.is_debug_enabled(self.logger)

    def init_class_attribute(self):
        self.logger.info("register managed attribute %s on class %s" % (self.key, self.parent.class_.__name__))
        sessionlib.attribute_manager.register_attribute(self.parent.class_, self.key, uselist=False, callable_=lambda i:self.setup_loader(i), copy_function=lambda x: self.columns[0].type.copy_value(x), compare_function=lambda x,y:self.columns[0].type.compare_values(x,y), mutable_scalars=self.columns[0].type.is_mutable())

    def setup_query(self, context, **kwargs):
        pass
        
    def process_row(self, selectcontext, instance, row, identitykey, isnew):
        if isnew:
            if not self.is_default or len(selectcontext.options):
                sessionlib.attribute_manager.init_instance_attribute(instance, self.key, False, callable_=self.setup_loader(instance, selectcontext.options))
            else:
                sessionlib.attribute_manager.reset_instance_attribute(instance, self.key)

    def setup_loader(self, instance, options=None):
        if not mapper.has_mapper(instance):
            return None
        else:
            prop = mapper.object_mapper(instance).props[self.key]
            if prop is not self.parent_property:
                return prop._get_strategy(DeferredColumnLoader).setup_loader(instance)
        def lazyload():
            if self._should_log_debug:
                self.logger.debug("deferred load %s group %s" % (mapperutil.attribute_str(instance, self.key), str(self.group)))

            if not mapper.has_identity(instance):
                return None

            try:
                pk = self.parent.pks_by_table[self.columns[0].table]
            except KeyError:
                pk = self.columns[0].table.primary_key

            clause = sql.and_()
            for primary_key in pk:
                attr = self.parent.get_attr_by_column(instance, primary_key)
                if not attr:
                    return None
                clause.clauses.append(primary_key == attr)

            session = sessionlib.object_session(instance)
            if session is None:
                raise exceptions.InvalidRequestError("Parent instance %s is not bound to a Session; deferred load operation of attribute '%s' cannot proceed" % (instance.__class__, self.key))

            localparent = mapper.object_mapper(instance)
            if self.group is not None:
                groupcols = [p for p in localparent.props.values() if isinstance(p.strategy, DeferredColumnLoader) and p.group==self.group]
                result = session.execute(localparent, sql.select([g.columns[0] for g in groupcols], clause, use_labels=True), None)
                try:
                    row = result.fetchone()
                    for prop in groupcols:
                        if prop is self:
                            continue
                        # set a scalar object instance directly on the object, 
                        # bypassing SmartProperty event handlers.
                        sessionlib.attribute_manager.init_instance_attribute(instance, prop.key, uselist=False)
                        instance.__dict__[prop.key] = row[prop.columns[0]]
                    return row[self.columns[0]]    
                finally:
                    result.close()
            else:
                return session.scalar(localparent, sql.select([self.columns[0]], clause, use_labels=True),None)

        return lazyload
                
DeferredColumnLoader.logger = logging.class_logger(DeferredColumnLoader)

class DeferredOption(StrategizedOption):
    def __init__(self, key, defer=False):
        super(DeferredOption, self).__init__(key)
        self.defer = defer

    def get_strategy_class(self):
        if self.defer:
            return DeferredColumnLoader
        else:
            return ColumnLoader

class AbstractRelationLoader(LoaderStrategy):
    def init(self):
        super(AbstractRelationLoader, self).init()
        for attr in ['primaryjoin', 'secondaryjoin', 'secondary', 'foreign_keys', 'mapper', 'select_mapper', 'target', 'select_table', 'loads_polymorphic', 'uselist', 'cascade', 'attributeext', 'order_by', 'remote_side', 'polymorphic_primaryjoin', 'polymorphic_secondaryjoin', 'direction']:
            setattr(self, attr, getattr(self.parent_property, attr))
        self._should_log_debug = logging.is_debug_enabled(self.logger)
        
    def _init_instance_attribute(self, instance, callable_=None):
        return sessionlib.attribute_manager.init_instance_attribute(instance, self.key, self.uselist, cascade=self.cascade,  trackparent=True, callable_=callable_)
        
    def _register_attribute(self, class_, callable_=None):
        self.logger.info("register managed %s attribute %s on class %s" % ((self.uselist and "list-holding" or "scalar"), self.key, self.parent.class_.__name__))
        sessionlib.attribute_manager.register_attribute(class_, self.key, uselist = self.uselist, extension=self.attributeext, cascade=self.cascade,  trackparent=True, typecallable=self.parent_property.collection_class, callable_=callable_)

class NoLoader(AbstractRelationLoader):
    def init_class_attribute(self):
        self._register_attribute(self.parent.class_)

    def process_row(self, selectcontext, instance, row, identitykey, isnew):
        if isnew:
            if not self.is_default or len(selectcontext.options):
                if self._should_log_debug:
                    self.logger.debug("set instance-level no loader on %s" % mapperutil.attribute_str(instance, self.key))
                self._init_instance_attribute(instance)

NoLoader.logger = logging.class_logger(NoLoader)
        
class LazyLoader(AbstractRelationLoader):
    def init(self):
        super(LazyLoader, self).init()
        (self.lazywhere, self.lazybinds, self.lazyreverse) = self._create_lazy_clause(self)
        
        self.logger.info(str(self.parent_property) + " lazy loading clause " + str(self.lazywhere))

        # determine if our "lazywhere" clause is the same as the mapper's
        # get() clause.  then we can just use mapper.get()
        self.use_get = not self.uselist and query.Query(self.mapper)._get_clause.compare(self.lazywhere)
        if self.use_get:
            self.logger.info(str(self.parent_property) + " will use query.get() to optimize instance loads")

    def init_class_attribute(self):
        self._register_attribute(self.parent.class_, callable_=lambda i: self.setup_loader(i))

    def setup_loader(self, instance, options=None):
        if not mapper.has_mapper(instance):
            return None
        else:
            prop = mapper.object_mapper(instance).props[self.key]
            if prop is not self.parent_property:
                return prop._get_strategy(LazyLoader).setup_loader(instance)
        def lazyload():
            self.logger.debug("lazy load attribute %s on instance %s" % (self.key, mapperutil.instance_str(instance)))
            params = {}
            allparams = True
            # if the instance wasnt loaded from the database, then it cannot lazy load
            # child items.  one reason for this is that a bi-directional relationship
            # will not update properly, since bi-directional uses lazy loading functions
            # in both directions, and this instance will not be present in the lazily-loaded
            # results of the other objects since its not in the database
            if not mapper.has_identity(instance):
                return None
            #print "setting up loader, lazywhere", str(self.lazywhere), "binds", self.lazybinds
            for col, bind in self.lazybinds.iteritems():
                params[bind.key] = self.parent.get_attr_by_column(instance, col)
                if params[bind.key] is None:
                    allparams = False
                    break

            if not allparams:
                return None

            session = sessionlib.object_session(instance)
            if session is None:
                try:
                    session = mapper.object_mapper(instance).get_session()
                except exceptions.InvalidRequestError:
                    raise exceptions.InvalidRequestError("Parent instance %s is not bound to a Session, and no contextual session is established; lazy load operation of attribute '%s' cannot proceed" % (instance.__class__, self.key))

            # if we have a simple straight-primary key load, use mapper.get()
            # to possibly save a DB round trip
            if self.use_get:
                ident = []
                for primary_key in self.select_mapper.pks_by_table[self.select_mapper.mapped_table]:
                    bind = self.lazyreverse[primary_key]
                    ident.append(params[bind.key])
                return session.query(self.mapper).get(ident)
            elif self.order_by is not False:
                order_by = self.order_by
            elif self.secondary is not None and self.secondary.default_order_by() is not None:
                order_by = self.secondary.default_order_by()
            else:
                order_by = False
            result = session.query(self.mapper, with_options=options).select_whereclause(self.lazywhere, order_by=order_by, params=params)

            if self.uselist:
                return result
            else:
                if len(result):
                    return result[0]
                else:
                    return None
        return lazyload

    def process_row(self, selectcontext, instance, row, identitykey, isnew):
        if isnew:
            # new object instance being loaded from a result row
            if not self.is_default or len(selectcontext.options):
                self.logger.debug("set instance-level lazy loader on %s" % mapperutil.attribute_str(instance, self.key))
                # we are not the primary manager for this attribute on this class - set up a per-instance lazyloader,
                # which will override the clareset_instance_attributess-level behavior
                self._init_instance_attribute(instance, callable_=self.setup_loader(instance, selectcontext.options))
            else:
                self.logger.debug("set class-level lazy loader on %s" % mapperutil.attribute_str(instance, self.key))
                # we are the primary manager for this attribute on this class - reset its per-instance attribute state, 
                # so that the class-level lazy loader is executed when next referenced on this instance.
                # this usually is not needed unless the constructor of the object referenced the attribute before we got 
                # to load data into it.
                sessionlib.attribute_manager.reset_instance_attribute(instance, self.key)

    def _create_lazy_clause(cls, prop, reverse_direction=False):
        (primaryjoin, secondaryjoin, remote_side) = (prop.polymorphic_primaryjoin, prop.polymorphic_secondaryjoin, prop.remote_side)
        
        binds = {}
        reverse = {}

        def should_bind(targetcol, othercol):
            if reverse_direction and not secondaryjoin:
                return targetcol in remote_side
            else:
                return othercol in remote_side

        def find_column_in_expr(expr):
            if not isinstance(expr, sql.ColumnElement):
                return None
            columns = []
            class FindColumnInColumnClause(sql.ClauseVisitor):
                def visit_column(self, c):
                    columns.append(c)
            FindColumnInColumnClause().traverse(expr)
            return len(columns) and columns[0] or None
        
        def bind_label():
            # TODO: make this generation deterministic
            return "lazy_" + hex(random.randint(0, 65535))[2:]

        def visit_binary(binary):
            leftcol = find_column_in_expr(binary.left)
            rightcol = find_column_in_expr(binary.right)
            if leftcol is None or rightcol is None:
                return
            if should_bind(leftcol, rightcol):
                col = leftcol
                binary.left = binds.setdefault(leftcol,
                        sql.bindparam(bind_label(), None, shortname=leftcol.name, type=binary.right.type, unique=True))
                reverse[rightcol] = binds[col]

            # the "left is not right" compare is to handle part of a join clause that is "table.c.col1==table.c.col1",
            # which can happen in rare cases (test/orm/relationships.py RelationTest2)
            if leftcol is not rightcol and should_bind(rightcol, leftcol):
                col = rightcol
                binary.right = binds.setdefault(rightcol,
                        sql.bindparam(bind_label(), None, shortname=rightcol.name, type=binary.left.type, unique=True))
                reverse[leftcol] = binds[col]

        lazywhere = primaryjoin.copy_container()
        li = mapperutil.BinaryVisitor(visit_binary)
        
        if not secondaryjoin or not reverse_direction:
            li.traverse(lazywhere)
        
        if secondaryjoin is not None:
            secondaryjoin = secondaryjoin.copy_container()
            if reverse_direction:
                li.traverse(secondaryjoin)
            lazywhere = sql.and_(lazywhere, secondaryjoin)

        return (lazywhere, binds, reverse)
    _create_lazy_clause = classmethod(_create_lazy_clause)
    
LazyLoader.logger = logging.class_logger(LazyLoader)


class EagerLoader(AbstractRelationLoader):
    """Loads related objects inline with a parent query."""
    
    def init(self):
        super(EagerLoader, self).init()
        if self.parent.isa(self.mapper):
            raise exceptions.ArgumentError(
                "Error creating eager relationship '%s' on parent class '%s' "
                "to child class '%s': Cant use eager loading on a self "
                "referential relationship." %
                (self.key, repr(self.parent.class_), repr(self.mapper.class_)))
        if self.is_default:
            self.parent._eager_loaders.add(self.parent_property)

        self.clauses = {}
        self.clauses_by_lead_mapper = {}

    class AliasedClauses(object):
        """Defines a set of join conditions and table aliases which
        are aliased on a randomly-generated alias name, corresponding
        to the connection of an optional parent AliasedClauses object
        and a target mapper.
        
        EagerLoader has a distinct AliasedClauses object per parent
        AliasedClauses object, so that all paths from one mapper to
        another across a chain of eagerloaders generates a distinct
        chain of joins.  The AliasedClauses objects are generated and
        cached on an as-needed basis.
        
        E.g.::
        
            mapper A -->
                (EagerLoader 'items') --> 
                    mapper B --> 
                        (EagerLoader 'keywords') --> 
                            mapper C
            
        will generate::
            
            EagerLoader 'items' --> {
                None : AliasedClauses(items, None, alias_suffix='AB34')        # mappera JOIN mapperb_AB34
            }
            
            EagerLoader 'keywords' --> [
                None : AliasedClauses(keywords, None, alias_suffix='43EF')     # mapperb JOIN mapperc_43EF
                AliasedClauses(items, None, alias_suffix='AB34') : 
                        AliasedClauses(keywords, items, alias_suffix='8F44')   # mapperb_AB34 JOIN mapperc_8F44
            ]
        """
        
        def __init__(self, eagerloader, parentclauses=None):
            self.id = (parentclauses is not None and (parentclauses.id + "/") or '') + str(eagerloader.parent_property)
            self.parent = eagerloader
            self.target = eagerloader.select_table
            self.eagertarget = eagerloader.select_table.alias(self._aliashash("/target"))
            self.extra_cols = {}

            if eagerloader.secondary:
                self.eagersecondary = eagerloader.secondary.alias(self._aliashash("/secondary"))
                if parentclauses is not None:
                    aliasizer = sql_util.ClauseAdapter(self.eagertarget).\
                            chain(sql_util.ClauseAdapter(self.eagersecondary)).\
                            chain(sql_util.ClauseAdapter(parentclauses.eagertarget))
                else:
                    aliasizer = sql_util.ClauseAdapter(self.eagertarget).\
                        chain(sql_util.ClauseAdapter(self.eagersecondary))
                self.eagersecondaryjoin = eagerloader.polymorphic_secondaryjoin.copy_container()
                aliasizer.traverse(self.eagersecondaryjoin)
                self.eagerprimary = eagerloader.polymorphic_primaryjoin.copy_container()
                aliasizer.traverse(self.eagerprimary)
            else:
                self.eagerprimary = eagerloader.polymorphic_primaryjoin.copy_container()
                if parentclauses is not None: 
                    aliasizer = sql_util.ClauseAdapter(self.eagertarget)
                    aliasizer.chain(sql_util.ClauseAdapter(parentclauses.eagertarget, exclude=eagerloader.parent_property.remote_side))
                else:
                    aliasizer = sql_util.ClauseAdapter(self.eagertarget)
                aliasizer.traverse(self.eagerprimary)

            if eagerloader.order_by:
                self.eager_order_by = sql_util.ClauseAdapter(self.eagertarget).copy_and_process(util.to_list(eagerloader.order_by))
            else:
                self.eager_order_by = None

            self._row_decorator = self._create_decorator_row()
        
        def aliased_column(self, column):
            """return the aliased version of the given column, creating a new label for it if not already
            present in this AliasedClauses eagertable."""

            conv = self.eagertarget.corresponding_column(column, raiseerr=False)
            if conv:
                return conv

            if column in self.extra_cols:
                return self.extra_cols[column]
            
            aliased_column = column.copy_container()
            sql_util.ClauseAdapter(self.eagertarget).traverse(aliased_column)
            alias = self._aliashash(column.name)
            aliased_column = aliased_column.label(alias)
            self._row_decorator.map[column] = alias
            self.extra_cols[column] = aliased_column
            return aliased_column
            
        def _aliashash(self, extra):
            """return a deterministic 4 digit hash value for this AliasedClause's id + extra."""
            # use the first 4 digits of an MD5 hash
            return "anon_" + util.hash(self.id + extra)[0:4]
            
        def _create_decorator_row(self):
            class EagerRowAdapter(object):
                def __init__(self, row):
                    self.row = row
                def has_key(self, key):
                    return map.has_key(key) or self.row.has_key(key)
                def __getitem__(self, key):
                    if map.has_key(key):
                        key = map[key]
                    return self.row[key]
                def keys(self):
                    return map.keys()
            map = {}        
            for c in self.eagertarget.c:
                parent = self.target.corresponding_column(c)
                map[parent] = c
                map[parent._label] = c
                map[parent.name] = c
            EagerRowAdapter.map = map
            return EagerRowAdapter

        def _decorate_row(self, row):
            # adapts a row at row iteration time to transparently
            # convert plain columns into the aliased columns that were actually
            # added to the column clause of the SELECT.
            return self._row_decorator(row)

    def init_class_attribute(self):
        self.parent_property._get_strategy(LazyLoader).init_class_attribute()
        
    def setup_query(self, context, eagertable=None, parentclauses=None, parentmapper=None, **kwargs):
        """Add a left outer join to the statement thats being constructed."""
        
        if parentmapper is None:
            localparent = context.mapper
        else:
            localparent = parentmapper
        
        if self.mapper in context.recursion_stack:
            return
        else:
            context.recursion_stack.add(self.parent)

        statement = context.statement
        
        if hasattr(statement, '_outerjoin'):
            towrap = statement._outerjoin
        elif isinstance(localparent.mapped_table, sql.Join):
            towrap = localparent.mapped_table
        else:
            # look for the mapper's selectable expressed within the current "from" criterion.
            # this will locate the selectable inside of any containers it may be a part of (such
            # as a join).  if its inside of a join, we want to outer join on that join, not the 
            # selectable.
            for fromclause in statement.froms:
                if fromclause is localparent.mapped_table:
                    towrap = fromclause
                    break
                elif isinstance(fromclause, sql.Join):
                    if localparent.mapped_table in sql_util.TableFinder(fromclause, include_aliases=True):
                        towrap = fromclause
                        break
            else:
                raise exceptions.InvalidRequestError("EagerLoader cannot locate a clause with which to outer join to, in query '%s' %s" % (str(statement), localparent.mapped_table))
        
        try:
            clauses = self.clauses[parentclauses]
        except KeyError:
            clauses = EagerLoader.AliasedClauses(self, parentclauses)
            self.clauses[parentclauses] = clauses
            
        if context.mapper not in self.clauses_by_lead_mapper:
            self.clauses_by_lead_mapper[context.mapper] = clauses

        if self.secondaryjoin is not None:
            statement._outerjoin = sql.outerjoin(towrap, clauses.eagersecondary, clauses.eagerprimary).outerjoin(clauses.eagertarget, clauses.eagersecondaryjoin)
            if self.order_by is False and self.secondary.default_order_by() is not None:
                statement.order_by(*clauses.eagersecondary.default_order_by())
        else:
            statement._outerjoin = towrap.outerjoin(clauses.eagertarget, clauses.eagerprimary)
            if self.order_by is False and clauses.eagertarget.default_order_by() is not None:
                statement.order_by(*clauses.eagertarget.default_order_by())

        if clauses.eager_order_by:
            statement.order_by(*util.to_list(clauses.eager_order_by))
                
        statement.append_from(statement._outerjoin)
        for value in self.select_mapper.props.values():
            value.setup(context, eagertable=clauses.eagertarget, parentclauses=clauses, parentmapper=self.select_mapper)

    def _create_row_processor(self, selectcontext, row):
        """Create a *row processing* function that will apply eager
        aliasing to the row.
        
        Also check that an identity key can be retrieved from the row,
        else return None.
        """
        
        # check for a user-defined decorator in the SelectContext (which was set up by the contains_eager() option)
        if selectcontext.attributes.has_key((EagerLoader, self.parent_property)):
            # custom row decoration function, placed in the selectcontext by the 
            # contains_eager() mapper option
            decorator = selectcontext.attributes[(EagerLoader, self.parent_property)]
            if decorator is None:
                decorator = lambda row: row
        else:
            try:
                # decorate the row according to the stored AliasedClauses for this eager load
                clauses = self.clauses_by_lead_mapper[selectcontext.mapper]
                decorator = clauses._row_decorator
            except KeyError, k:
                # no stored AliasedClauses: eager loading was not set up in the query and
                # AliasedClauses never got initialized
                return None

        try:
            decorated_row = decorator(row)
            # check for identity key
            identity_key = self.mapper.identity_key_from_row(decorated_row)
            # and its good
            return decorator
        except KeyError, k:
            # no identity key - dont return a row processor, will cause a degrade to lazy
            if self._should_log_debug:
                self.logger.debug("could not locate identity key from row '%s'; missing column '%s'" % (repr(decorated_row), str(k)))
            return None

    def process_row(self, selectcontext, instance, row, identitykey, isnew):
        """Receive a row.

        Tell our mapper to look for a new object instance in the row,
        and attach it to a list on the parent instance.
        """
        
        if self in selectcontext.recursion_stack:
            return
        
        try:
            # check for row processor
            row_processor = selectcontext.attributes[id(self)]
        except KeyError:
            # create a row processor function and cache it in the context
            row_processor = self._create_row_processor(selectcontext, row)
            selectcontext.attributes[id(self)] = row_processor
            
        if row_processor is not None:
            decorated_row = row_processor(row)
        else:
            # row_processor was None: degrade to a lazy loader
            if self._should_log_debug:
                self.logger.debug("degrade to lazy loader on %s" % mapperutil.attribute_str(instance, self.key))
            self.parent_property._get_strategy(LazyLoader).process_row(selectcontext, instance, row, identitykey, isnew)
            return
            
        # TODO: recursion check a speed hit...?  try to get a "termination point" into the AliasedClauses
        # or EagerRowAdapter ?
        selectcontext.recursion_stack.add(self)
        try:
            if not self.uselist:
                if self._should_log_debug:
                    self.logger.debug("eagerload scalar instance on %s" % mapperutil.attribute_str(instance, self.key))
                if isnew:
                    # set a scalar object instance directly on the parent object, 
                    # bypassing SmartProperty event handlers.
                    instance.__dict__[self.key] = self.mapper._instance(selectcontext, decorated_row, None)
                else:
                    # call _instance on the row, even though the object has been created,
                    # so that we further descend into properties
                    self.mapper._instance(selectcontext, decorated_row, None)
            else:
                if isnew:
                    if self._should_log_debug:
                        self.logger.debug("initialize UniqueAppender on %s" % mapperutil.attribute_str(instance, self.key))
                    # call the SmartProperty's initialize() method to create a new, blank list
                    l = getattr(instance.__class__, self.key).initialize(instance)
                
                    # create an appender object which will add set-like semantics to the list
                    appender = util.UniqueAppender(l.data)
                
                    # store it in the "scratch" area, which is local to this load operation.
                    selectcontext.attributes[(instance, self.key)] = appender
                result_list = selectcontext.attributes[(instance, self.key)]
                if self._should_log_debug:
                    self.logger.debug("eagerload list instance on %s" % mapperutil.attribute_str(instance, self.key))
                self.select_mapper._instance(selectcontext, decorated_row, result_list)
        finally:
            selectcontext.recursion_stack.remove(self)

EagerLoader.logger = logging.class_logger(EagerLoader)

class EagerLazyOption(StrategizedOption):
    def __init__(self, key, lazy=True):
        super(EagerLazyOption, self).__init__(key)
        self.lazy = lazy

    def process_query_property(self, context, prop):
        if self.lazy:
            if prop in context.eager_loaders:
                context.eager_loaders.remove(prop)
        else:
            context.eager_loaders.add(prop)
        super(EagerLazyOption, self).process_query_property(context, prop)

    def get_strategy_class(self):
        if self.lazy:
            return LazyLoader
        elif self.lazy is False:
            return EagerLoader
        elif self.lazy is None:
            return NoLoader

EagerLazyOption.logger = logging.class_logger(EagerLazyOption)

class RowDecorateOption(PropertyOption):
    def __init__(self, key, decorator=None, alias=None):
        super(RowDecorateOption, self).__init__(key)
        self.decorator = decorator
        self.alias = alias

    def process_selection_property(self, context, property):
        if self.alias is not None and self.decorator is None:
            if isinstance(self.alias, basestring):
                self.alias = property.target.alias(self.alias)
            def decorate(row):
                d = {}
                for c in property.target.columns:
                    d[c] = row[self.alias.corresponding_column(c)]
                return d
            self.decorator = decorate
        context.attributes[(EagerLoader, property)] = self.decorator

RowDecorateOption.logger = logging.class_logger(RowDecorateOption)
        




# -*- coding: UTF-8 -*-

__revision__ = '$Id$'

# Copyright (c) 2006-2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

import gutils
import movie
import string
import re
import amazon
import gdebug
import threading
import gtk
from operator import isSequenceType
from urlparse import urlsplit

plugin_name = "Amazon"
plugin_description = "Amazon"
plugin_url = "www.amazon.com/.uk/.de/.ca/.fr/.jp"
plugin_language = _("International")
plugin_author = "Michael Jahn"
plugin_author_email = "<mikej06@hotmail.com>"
plugin_version = "1.0"

class Plugin(movie.Movie):

	def __init__(self, id):
		self.encode='utf8'
		self.movie_id = id
		self.url = 'http://www.amazon.de/dp/' + str(self.movie_id)

	def open_page(self, parent_window=None, url=None):
		# dont use base functionality
		# use the Amazon Web API
		self.parent_window = parent_window
		try:
			locale = self.config.get('amazon_locale', 0, section='add')
			if locale == '1':
				locale = 'uk'
			elif locale == '2':
				locale = 'de'
			elif locale == '3':
				locale = 'ca'
			elif locale == '4':
				locale = 'fr'
			elif locale == '5':
				locale = 'jp'
			else:
				locale = None
			retriever = AmazonRetriever(self.movie_id, locale, parent_window, self.progress, self.debug, 'Get')
			retriever.start()
			while retriever.isAlive():
				self.progress.pulse()
				while gtk.events_pending():
					gtk.main_iteration()
			self.page = retriever.result.Item
		except:
			self.page = ''
			try:
				self.debug.show("Error retrieving results from amazon.")
				self.debug.show(retriever.result.Request.Errors.Error.Message)
			except:
				pass
		return self.page

	def get_image(self):
		self.image_url = ''
		if hasattr(self.page, 'LargeImage'):
			self.image_url = self.page.LargeImage.URL
		elif hasattr(self.page, 'MediumImage'):
			self.image_url = self.page.MediumImage.URL
		elif hasattr(self.page, 'SmallImage'):
			self.image_url = self.page.SmallImage.URL

	def get_o_title(self):
		if hasattr(self.page.ItemAttributes, 'Title'):
			self.o_title = self.page.ItemAttributes.Title
		else:
			self.director = ''

	def get_title(self):
		if hasattr(self.page.ItemAttributes, 'Title'):
			self.title = self.page.ItemAttributes.Title
		else:
			self.director = ''

	def get_director(self):
		if hasattr(self.page.ItemAttributes, 'Director'):
			self.director = self.page.ItemAttributes.Director
		else:
			self.director = ''

	def get_plot(self):
		self.plot = ''
		if hasattr(self.page, 'EditorialReviews'):
			if hasattr(self.page.EditorialReviews, 'EditorialReview'):
				if isSequenceType(self.page.EditorialReviews.EditorialReview):
					for review in self.page.EditorialReviews.EditorialReview:
						if string.find(review.Source, 'Amazon') > -1:
							self.plot = review.Content
				else:
					if hasattr(self.page.EditorialReviews.EditorialReview, 'Source') and \
						hasattr(self.page.EditorialReviews.EditorialReview, 'Content') and \
						string.find(self.page.EditorialReviews.EditorialReview.Source, 'Amazon') > -1:
						self.plot = self.page.EditorialReviews.EditorialReview.Content

	def get_year(self):
		if hasattr(self.page.ItemAttributes, 'TheatricalReleaseDate'):
			self.year = self.page.ItemAttributes.TheatricalReleaseDate[:4]
		elif hasattr(self.page.ItemAttributes, 'ReleaseDate'):
			self.year = self.page.ItemAttributes.ReleaseDate[:4]
		else:
			self.year = ''

	def get_runtime(self):
		if hasattr(self.page.ItemAttributes, 'RunningTime'):
			self.runtime = self.page.ItemAttributes.RunningTime
		else:
			self.runtime = ''

	def get_genre(self):
		# BrowseNodeId 547664 (Genres)
		self.genre = ''
		delimiter = ''
		if hasattr(self.page, 'BrowseNodes') and hasattr(self.page.BrowseNodes, 'BrowseNode'):
			if isSequenceType(self.page.BrowseNodes.BrowseNode):
				for node in self.page.BrowseNodes.BrowseNode:
					parentnode = node
					while hasattr(parentnode, 'Ancestors') and parentnode.BrowseNodeId <> '547664' \
							and parentnode.BrowseNodeId <> '13628901': # no production countries; they are also arranged under genres
						parentnode = parentnode.Ancestors.BrowseNode
					if parentnode.BrowseNodeId == '547664':
						self.genre = self.genre + delimiter + node.Name
						delimiter = ', '

	def get_cast(self):
		self.cast = ''
		if hasattr(self.page.ItemAttributes, 'Actor'):
			for actor in self.page.ItemAttributes.Actor:
				self.cast += actor + '\n'

	def get_classification(self):
		if hasattr(self.page.ItemAttributes, 'AudienceRating'):
			self.classification = self.page.ItemAttributes.AudienceRating
		else:
			self.classification = ''

	def get_studio(self):
		if hasattr(self.page.ItemAttributes, 'Studio'):
			self.studio = self.page.ItemAttributes.Studio
		else:
			self.studio = ''

	def get_o_site(self):
		self.o_site = ''

	def get_site(self):
		if hasattr(self.page, 'DetailPageURL'):
			parts = urlsplit(self.page.DetailPageURL)
			self.site = parts[0] + '://' + parts[1] + '/dp/' + self.movie_id
		else:
			self.site = ''

	def get_trailer(self):
		self.trailer = ''

	def get_country(self):
		# BrowseNodeId 13628901 (production countries)
		self.country = ''
		delimiter = ''
		if hasattr(self.page, 'BrowseNodes') and hasattr(self.page.BrowseNodes, 'BrowseNode'):
			if isSequenceType(self.page.BrowseNodes.BrowseNode):
				for node in self.page.BrowseNodes.BrowseNode:
					parentnode = node
					while hasattr(parentnode, 'Ancestors') and parentnode.BrowseNodeId <> '13628901':
						parentnode = parentnode.Ancestors.BrowseNode
					if parentnode.BrowseNodeId == '13628901':
						self.country = self.country + delimiter + node.Name
						delimiter = ', '

	def get_rating(self):
		self.rating = '0'
		if hasattr(self.page, 'CustomerReviews') and \
			hasattr(self.page.CustomerReviews, 'AverageRating'):
			try:
				tmp_float = float(self.page.CustomerReviews.AverageRating)
				tmp_float = round(2 * tmp_float, 0)
				self.rating = str(tmp_float)
			except:
				pass

	def get_notes(self):
		self.notes = ''
		if hasattr(self.page.ItemAttributes, 'EAN'):
			self.notes = 'EAN: ' + self.page.ItemAttributes.EAN


class SearchPlugin(movie.SearchMovie):

	def __init__(self):
		self.original_url_search   = 'http://www.amazon.de'
		self.translated_url_search = 'http://www.amazon.de'
		self.encode='utf8'
		self.remove_accents = False

	def search(self,parent_window):
		# dont use base functionality
		# use the Amazon Web API
		self.titles = [""]
		self.ids = [""]
		try:
			locale = self.config.get('amazon_locale', 0, section='add')
			if locale == '1':
				locale = 'uk'
			elif locale == '2':
				locale = 'de'
			elif locale == '3':
				locale = 'ca'
			elif locale == '4':
				locale = 'fr'
			elif locale == '5':
				locale = 'jp'
			else:
				locale = None
			retriever = AmazonRetriever(self.title.encode('iso8859-1'), locale, parent_window, self.progress, self.debug)
			retriever.start()
			while retriever.isAlive():
				self.progress.pulse()
				while gtk.events_pending():
					gtk.main_iteration()
			self.page = retriever.result
		except:
			try:
				self.debug.show("Error retrieving results from amazon.")
				self.debug.show(retriever.result.Request.Errors.Error.Message)
			except:
				pass
		return self.page

	def get_searches(self):
		for result in self.page:
			if hasattr(result, 'Item'):
				if hasattr(result.Item, 'ASIN'):
					self.add_item(result.Item)
				else:
					for item in result.Item:
						self.add_item(item)
			elif hasattr(result, 'Items'):
				if hasattr(result.Item, 'ASIN'):
					self.add_item(result.Items)
				else:
					for item in result.Items:
						self.add_item(item)

	def add_item(self, item):
		self.ids.append(item.ASIN)
		if hasattr(item.ItemAttributes, 'ProductGroup'):
			productGroup = item.ItemAttributes.ProductGroup + ' - '
		else:
			productGroup = ''
		if hasattr(item.ItemAttributes, 'Title'):
			title = item.ItemAttributes.Title
		else:
			title = ''
		if hasattr(item.ItemAttributes, 'TheatricalReleaseDate'):
			theatricalReleaseDate = ' (' + item.ItemAttributes.TheatricalReleaseDate + ')'
		else:
			theatricalReleaseDate = ''
		self.titles.append("%s%s%s" % (productGroup, title, theatricalReleaseDate))

class AmazonRetriever(threading.Thread):

	def __init__(self, title, locale, parent_window, progress, debug, lookuptype='Search', destination=None):
		self.title = title
		self.locale = locale
		self.result = None
		self.destination = destination
		self.parent_window = parent_window
		self.progress = progress
		self.debug = debug
		# Search or Get
		self.lookuptype = lookuptype
		self._stopevent = threading.Event()
		self._sleepperiod = 1.0
		threading.Thread.__init__(self, name='Retriever')

	def run(self):
		if self.lookuptype == 'Get':
			self.run_get()
		else:
			self.run_search()

	def run_search(self):
		self.result = []
		try:
			amazon.setLicense('04GDDMMXX8X9CJ1B22G2')
			try:
				tmp = amazon.searchByTitle(self.title, type='ItemAttributes', product_line='Video', locale=self.locale, page=1)
				self.result.append(tmp)
				if hasattr(tmp, 'TotalPages'):
					pages = int(tmp.TotalPages) - 1
					page = 2
					while page < pages and page < 11:
						tmp = amazon.searchByTitle(self.title, type='ItemAttributes', product_line='Video', locale=self.locale, page=page)
						self.result.append(tmp)
						page = page + 1
			except amazon.AmazonError, e:
				self.debug.show(e.Message)
			# if all digits then try to find an EAN / UPC
			if self.title.isdigit():
				if len(self.title) == 13:
					try:
						tmp = amazon.searchByEAN(self.title, type='ItemAttributes', product_line='Video', locale=self.locale)
						self.result.append(tmp)
					except amazon.AmazonError, e:
						self.debug.show(e.Message)
				elif len(self.title) == 12:
					try:
						tmp = amazon.searchByUPC(self.title, type='ItemAttributes', product_line='Video', locale=self.locale)
						self.result.append(tmp)
					except amazon.AmazonError, e:
						self.debug.show(e.Message)
		except IOError:
			self.progress.dialog.hide()
			gutils.urllib_error(_('Connection error'), self.parent_window)
			self.suspend()

	def run_get(self):
		self.result = None
		try:
			amazon.setLicense('04GDDMMXX8X9CJ1B22G2')
			# get by ASIN
			try:
				self.result = amazon.searchByASIN(self.title, type='Large', locale=self.locale)
			except amazon.AmazonError, e:
				self.debug.show(e.Message)
		except IOError:
			self.progress.dialog.hide()
			gutils.urllib_error(_('Connection error'), self.parent_window)
			self.suspend()

#
# Plugin Test
#
class SearchPluginTest(SearchPlugin):
	#
	# Configuration for automated tests:
	# dict { movie_id -> [ expected result count for original url, expected result count for translated url ] }
	#
	test_configuration = {
		'Rocky Balboa'			: [ 10, 10 ],
		'Arahan'				: [ 6, 6 ],
		'Ein glückliches Jahr'	: [ 2, 2 ]
	}

class PluginTest:
	#
	# Configuration for automated tests:
	# dict { movie_id -> dict { arribute -> value } }
	#
	# value: * True/False if attribute only should be tested for any value
	#        * or the expected value
	#
	test_configuration = {
		'B000TIQMMI' : { 
			'title' 			: 'Rocky Balboa',
			'o_title' 			: 'Rocky Balboa',
			'director'			: '',
			'plot' 				: True,
			'cast'				: 'Sylvester Stallone\n\
Antonio Traver\n\
Burt Young',
			'country'			: 'USA',
			'genre'				: 'Mehr Drama, Drama',
			'classification'	: 'Freigegeben ab 12 Jahren',
			'studio'			: 'MGM Home Entertainment GmbH (dt.)',
			'o_site'			: False,
			'site'				: 'http://www.amazon.de/dp/B000TIQMMI',
			'trailer'			: False,
			'year'				: 2006,
			'notes'				: 'EAN: 4045167004504',
			'runtime'			: 97,
			'image'				: True,
			'rating'			: 9
		},
		'B0009NSASM' : { 
			'title' 			: 'Ein glückliches Jahr',
			'o_title' 			: 'Ein glückliches Jahr',
			'director'			: 'Claude Lelouch',
			'plot' 				: False,
			'cast'				: 'Lino Ventura\n\
Françoise Fabian\n\
Charles Gérard',
			'country'			: 'Frankreich, Italien',
			'genre'				: 'Krimi, Mehr Drama, Mehr Komödie, Drama, Komödie, Krimi',
			'classification'	: 'Freigegeben ab 12 Jahren',
			'studio'			: 'Warner Home Video - DVD',
			'o_site'			: False,
			'site'				: 'http://www.amazon.de/dp/B0009NSASM',
			'trailer'			: False,
			'year'				: 1973,
			'notes'				: 'EAN: 7321921998843',
			'runtime'			: 110,
			'image'				: True,
			'rating'			: 10
		},
		'B000BSNOD6' : { 
			'title' 			: 'Arahan (Vanilla-DVD)',
			'o_title' 			: 'Arahan (Vanilla-DVD)',
			'director'			: 'Ryoo Seung-wan',
			'plot' 				: False,
			'cast'				: 'Ryu Seung-beom\n\
Yoon So-yi\n\
Ahn Sung-kee',
			'country'			: 'Südkorea',
			'genre'				: 'Actionkomödie, Abenteuer- & Actionkomödie, Fantasykomödie, Action, Komödie, Mehr Fantasy, Korea, Action, Komödie, Fantasy, Unter 10 EUR',
			'classification'	: 'Freigegeben ab 16 Jahren',
			'o_site'			: False,
			'site'				: 'http://www.amazon.de/dp/B000BSNOD6',
			'trailer'			: False,
			'year'				: 2005,
			'notes'				: 'EAN: 4013549871105',
			'runtime'			: 108,
			'image'				: True,
			'rating'			: 8
		}
	}



# -*- coding: UTF-8 -*-

__revision__ = '$Id: backup.py 1040 2008-11-15 21:13:49Z mikej06 $'

# Copyright (c) 2005-2008 Vasco Nunes, Piotr Ożarowski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

import config, edit, gutils, sql
import gtk
import os.path
import zipfile

def backup(self):
	"""perform a compressed griffith database/posters/preferences backup"""
	filename = gutils.file_chooser(_("Save Griffith backup"), \
		action=gtk.FILE_CHOOSER_ACTION_SAVE, buttons= \
		(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_SAVE,gtk.RESPONSE_OK), \
		name=self.config.get('name','griffith', section='database') + '_backup.zip')
	if filename and filename[0]:
		overwrite = None
		zipfilename = filename[0].decode('utf-8')
		if os.path.isfile(zipfilename):
			response = gutils.question(self, \
				_("File exists. Do you want to overwrite it?"), \
				1, self.widgets['window'])
			if response == -8:
				overwrite = True
			else:
				overwrite = False

		if overwrite == True or overwrite is None:
			try:
				if zipfile.zlib is not None:
					mzip = zipfile.ZipFile(zipfilename, 'w', zipfile.ZIP_DEFLATED)
				else:
					mzip = zipfile.ZipFile(zipfilename, 'w')
			except:
				gutils.error(self, _("Error creating backup"), self.widgets['window'])
				return False
			mzip.write(os.path.join(self.locations['home'],'griffith.cfg').encode('utf-8'))
			if self.db.metadata.engine.name == 'sqlite':
				fileName = os.path.join(self.locations['home'], self.config.get('name','griffith', section='database') + '.db').encode('utf-8')
				mzip.write(fileName)
			else:
				gutils.error(self, _("Backup function is available only for SQLite engine for now"), self.widgets['window'])
				return False
				from tempfile import mkdtemp
				from shutil import rmtree, move
				from sqlalchemy import BoundMetaData
				import copy
				# if backup_to_sqlite:
				tmp_dir = mkdtemp()
				tmp_config = copy.deepcopy(self.config)
				tmp_config._file = os.path.join(tmp_dir,'griffith.cfg')
				tmp_config.get('type', 'sqlite', section='database') == 'sqlite'
				tmp_config.set('file', section='database') == "griffith.db"
				tmp_config.save()

#				tmp_db = sql.GriffithSQL(tmp_config, self.debug, tmp_dir)
#				for i in self.db.metadata.tables
#				tmp_db.

				tmp_file = os.path.join(tmp_dir, tmp_config.get('file', 'griffith.db', section='database')).encode('utf-8')
				tmp_metadata = BoundMetaData("sqlite:///%s" % tmp_file)
				tmp_metadata.tables = self.db.metadata.tables
				tmp_metadata.create_all()
#				for table in self.db.metadata.tables.keys():
				for table in [t.name for t in self.db.metadata.table_iterator()]: # table_iterator() will return tables in *correct* order
					data = self.db.metadata.tables[table].select().execute().fetchall()
					tmp_metadata.tables[table].insert().execute(data)
#					for item in data:
#						tmp_metadata.tables[table].insert().execute(item)
				tmp_metadata.engine.commit()
				
				mzip.write(tmp_file)
				rmtree(tmp_dir)
			posters_dir = os.path.join(self.locations['posters'])
			for movie in self.db.Movie.select():
				if movie.image is not None:
					filename = str(movie.image)+".jpg"
					filename = os.path.join(posters_dir, filename).encode('utf-8')
					if os.path.isfile(filename):
						try:
							mzip.write(filename)
						except:
							self.debug.show("Can't compress %s" % filename)
			mzip.close()
			gutils.info(self, _("Backup has been created"), self.widgets['window'])

def restore(self):
	"""restores a griffith compressed backup"""
	filename = gutils.file_chooser(_("Restore Griffith backup"), \
		action=gtk.FILE_CHOOSER_ACTION_OPEN, buttons= \
		(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, \
		gtk.STOCK_OPEN, gtk.RESPONSE_OK))
	if filename[0]:
		try:
			zip = zipfile.ZipFile(filename[0], 'r')
		except:
			gutils.error(self, _("Can't read backup file"), self.widgets['window'])
			return False
		mypath = os.path.join(self.locations['posters'])
		old_config_file = False
		for each in zip.namelist():
			file_to_restore = os.path.split(each)
			if not os.path.isdir(file_to_restore[1]):
				if file_to_restore[1] == '':
					continue
				if file_to_restore[1].endswith('.jpg'):
					myfile = os.path.join(mypath,file_to_restore[1])
				else:
					myfile = os.path.join(self.locations['home'],file_to_restore[1])
				if file_to_restore[1].endswith('.conf'):
					old_config_file = myfile
				outfile = open(myfile, 'wb')
				outfile.write(zip.read(each))
				outfile.flush()
				outfile.close()
		zip.close()

		# restore config file
		self.config = config.Config(file=os.path.join(self.locations['home'],'griffith.cfg'))
		if old_config_file:
			self.debug.show('Old config file detected. Please note that it will not be used.')
			f = open(old_config_file, 'r')
			old_config_raw_data = f.read()
			f.close()
			if old_config_raw_data.find('griffith.gri') >= -1:
				self.config.set('file', 'griffith.gri', section='database')

		filename = os.path.join(self.locations['home'], self.config.get('name', 'griffith', section='database') + '.db')

		self.db.metadata.engine.dispose() # close DB
		from sqlalchemy.orm import clear_mappers
		clear_mappers()

		# check if file needs conversion
		if self.config.get('file', 'griffith.db', section='database').lower().endswith('.gri'):
			self.debug.show('Old database format detected. Converting...')
			from dbupgrade import convert_from_old_db
			from initialize	import location_posters
			if convert_from_old_db(self, filename, os.path.join(self.locations['home'], 'griffith.db')):
				self.config.save()
				location_posters(self.locations, self.config)
			else:
				print 'Cant convert old database, exiting.'
				import sys
				sys.exit(4)

		self.db = sql.GriffithSQL(self.config, self.debug, self.locations['home'])
		from initialize	import dictionaries, people_treeview
		dictionaries(self)
		people_treeview(self)
		# let's refresh the treeview
		self.clear_details()
		self.populate_treeview()
		gutils.info(self, _("Backup restored"), self.widgets['window'])

def merge(self):	# FIXME
	"""
		Merge database from:
		* compressed backup
		* SQLite2 *.gri file
		* SQLite3 *.db file
	"""
	filename = gutils.file_chooser(_("Restore Griffith backup"), \
		action=gtk.FILE_CHOOSER_ACTION_OPEN, buttons= \
		(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, \
		gtk.STOCK_OPEN, gtk.RESPONSE_OK))[0]
	if filename:
		from tempfile import mkdtemp
		from shutil import rmtree, move

		#tmp_config={}
		#tmp_config.get('type', 'sqlite', section='database')

		if filename.lower().endswith('.zip'):
			tmp_dir = mkdtemp()
			try:
				zip = zipfile.ZipFile(filename, 'r')
			except:
				gutils.error(self, _("Can't read backup file"), self.widgets['window'])
				return False
			for each in zip.namelist():
				file_to_restore = os.path.split(each)
				if not os.path.isdir(file_to_restore[1]):
					myfile = os.path.join(tmp_dir, file_to_restore[1])
					outfile = open(myfile, 'wb')
					outfile.write(zip.read(each))
					outfile.flush()
					outfile.close()
			# load stored database filename
			tmp_config = config.Config(file=os.path.join(tmp_dir,'griffith.conf'))
			filename = os.path.join(tmp_dir, tmp_config('name', 'griffith', section='database') + '.db')
			zip.close()

		# check if file needs conversion
		if filename.lower().endswith(".gri"):
			if os.path.isfile(filename) and  open(filename).readline()[:47] == "** This file contains an SQLite 2.1 database **":
				self.debug.show("MERGE: SQLite2 database format detected. Converting...")
				if not self.db.convert_from_sqlite2(filename, os.path.join(tmp_dir, self.config.get('file', 'griffith.db', section='database'))):
					self.debug.show("MERGE: Can't convert database, aborting.")
					return False
		tmp_dir, tmp_file = os.path.split(filename)
		self.config.get('file', tmp_file, section='database') 

		tmp_db = sql.GriffithSQL(tmp_config, self.debug, tmp_dir)

		merged=0
		movies = tmp_db.Movie.count()
		for movie in tmp_db.Movie.select():
			if self.db.Movie.get_by(o_title=movie.o_title) is not None:
				continue
			t_movies = {}
			for column in movie.mapper.c.keys():
				t_movies[column] = eval("movie.%s"%column)

			# replace number with new one
			t_movies["number"] = gutils.find_next_available(self.db)

			# don't restore volume/collection/tag/language/loan data (it's dangerous)
			t_movies.pop('movie_id')
			t_movies.pop('loaned')
			t_movies.pop('volume_id')
			t_movies.pop('collection_id')

			if self.db.add_movie(t_movies):
				print t_movies

			if movie.image is not None:
				dest_file = os.path.join(self.locations['posters'], movie.image+'.jpg')
				if not os.path.isfile(dest_file):
					src_file = os.path.join(tmp_dir, movie.image+'.jpg')
					if os.path.isfile(src_file):
						move(src_file, dest_file)
			merged+=1
		rmtree(tmp_dir)

		from initialize	import dictionaries, people_treeview
		dictionaries(self)
		people_treeview(self)
		# let's refresh the treeview
		self.clear_details()
		self.populate_treeview(self.db.Movie.select())
		#gutils.info(self, _("Databases merged!\n\nProcessed movies: %s\nMerged movies: %s"%(movies, merged)), self.widgets['window'])




# -*- coding: UTF-8 -*-

__revision__ = '$Id: movie.py 1040 2008-11-15 21:13:49Z mikej06 $'

# Copyright (c) 2005-2008 Vasco Nunes, Piotr Ożarowski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

from urllib import *
import sys
import string
import gutils
import gtk
import os
import os.path
import threading
import time
import tempfile

class Movie:
	cast = None
	classification = None
	country = None
	director = None
	genre = None
	image = None
	notes = None
	number = None
	o_site = None
	o_title = None
	plot = None
	rating = None
	runtime = None
	site = None
	studio = None
	title = None
	trailer = None
	year = None
	
	movie_id = None
	debug = False
	locations = None
	engine_author = None
	engine_description = None
	engine_language = None
	engine_name = None
	engine_version = None
	page = None
	url = None
	image_url = None
	encode = 'iso-8859-1'
	fields_to_fetch = []
	progress = None

	# functions that plugin should implement: {{{
	def initialize(self):
		pass
	def get_cast(self):
		pass
	def get_classification(self):
		pass
	def get_country(self):
		pass
	def get_director(self):
		pass
	def get_genre(self):
		pass
	def get_image(self):
		pass
	def get_notes(self):
		pass
	def get_o_site(self):
		pass
	def get_o_title(self):
		pass
	def get_plot(self):
		pass
	def get_rating(self):
		pass
	def get_runtime(self):
		pass
	def get_site(self):
		pass
	def get_studio(self):
		pass
	def get_title(self):
		pass
	def get_trailer(self):
		pass
	def get_year(self):
		pass
	#}}}

	def __getitem__(self, key):
		return getattr(self,key)
	def __setitem__(self, key, value):
		setattr(self,key,value)
	
	def get_movie(self, parent_window=None):
		try:
			#
			# initialize the progress dialog once for the following loading process
			#
			if self.progress is None:
				self.progress = Progress(parent_window)
			self.progress.set_data(parent_window, _("Fetching data"), _("Wait a moment"), True)
			#			
			# get the page
			#
			self.open_page(parent_window)
		except:
			# close the dialog if an error occured
			self.progress.hide()
			# reraise the error
			raise

	def open_page(self, parent_window=None, url=None):
		if url is None:
			url_to_fetch = self.url
		else:
			url_to_fetch = url
		if parent_window is not None:
			self.parent_window = parent_window
		self.progress.set_data(parent_window, _("Fetching data"), _("Wait a moment"), False)
		retriever = Retriever(url_to_fetch, self.parent_window, self.progress)
		retriever.start()
		while retriever.isAlive():
			self.progress.pulse()
			if self.progress.status:
				retriever.suspend()
			while gtk.events_pending():
				gtk.main_iteration()
		try:
			ifile = file(retriever.html[0], "rb")
			data = ifile.read()
		except IOError:
			pass
		if url is None:
			self.page = data
		return data
		urlcleanup()

	def fetch_picture(self):
		if self.image_url:
			tmp_dest = tempfile.mktemp(prefix='poster_', dir=self.locations['temp'])
			self.image = tmp_dest.split('poster_', 1)[1]
			dest = "%s.jpg" % tmp_dest
			try:
				self.progress.set_data(self.parent_window, _("Fetching poster"), _("Wait a moment"), False)
				retriever = Retriever(self.image_url, self.parent_window, self.progress, dest)
				retriever.start()
				while retriever.isAlive():
					self.progress.pulse()
					if self.progress.status:
						retriever.suspend()
					while gtk.events_pending():
						gtk.main_iteration()
				urlcleanup()
			except:
				self.image = ""
				try:
					os.remove("%s.jpg" % tmp_dest )
				except:
					print "Can't remove %s file" % tmp_dest # FIXME: use debug.show()
		else:
			self.image = ""

	def parse_movie(self):
		try:
			from copy import deepcopy
			fields = deepcopy(self.fields_to_fetch)

			self.initialize()

			if 'year' in fields:
				self.get_year()
				self.year = gutils.digits_only(self.year, 2100)
				fields.pop(fields.index('year'))
			if 'runtime' in fields:
				self.get_runtime()
				self.runtime = gutils.digits_only(self.runtime)
				fields.pop(fields.index('runtime'))
			if 'rating' in fields:
				self.get_rating()
				self.rating = gutils.digits_only(self.rating, 10)
				fields.pop(fields.index('rating'))
			if 'cast' in fields:
				self.get_cast()
				self.cast = gutils.clean(self.cast)
				self.cast = gutils.gdecode(self.cast, self.encode)
				fields.pop(fields.index('cast'))
			if 'plot' in fields:
				self.get_plot()
				self.plot = gutils.clean(self.plot)
				self.plot = gutils.gdecode(self.plot, self.encode)
				fields.pop(fields.index('plot'))
			if 'notes' in fields:
				self.get_notes()
				self.notes = gutils.clean(self.notes)
				self.notes = gutils.gdecode(self.notes, self.encode)
				fields.pop(fields.index('notes'))
			if 'image' in fields:
				self.get_image()
				self.fetch_picture()
				fields.pop(fields.index('image'))

			for i in fields:
				getattr(self, "get_%s" % i)()
				self[i] = gutils.clean(self[i])
				self[i] = gutils.gdecode(self[i], self.encode)
		
			if 'o_title' in self.fields_to_fetch and self.o_title is not None:
				if self.o_title[:4] == 'The ':
					self.o_title = self.o_title[4:] + ', The'
			if 'title' in self.fields_to_fetch and self.title is not None:
				if self.title[:4] == 'The ':
					self.title = self.title[4:] + ', The'
		finally:
			# close the progress dialog which was opened in get_movie
			self.progress.hide()

class SearchMovie:
	page = None
	number_results = None
	titles = [""]
	ids = [""]
	url = None
	encode = 'utf-8'
	original_url_search = None
	translated_url_search = None
	elements = None
	title = None
	remove_accents = True
	progress = None

	def __init__(self):
		pass

	def search_movies(self,parent_window):
		try:
			#
			# initialize the progress dialog once for the following search process
			#
			if self.progress is None:
				self.progress = Progress(parent_window)
			self.progress.set_data(parent_window, _("Searching"), _("Wait a moment"), True)
			#			
			# call the plugin specific search method
			#
			self.search(parent_window)
		finally:
			# dont forget to hide the progress dialog
			self.progress.hide()

	def open_search(self,parent_window):
		self.titles = [""]
		self.ids = [""]
		if self.url.find('%s') > 0:
			self.url = self.url % self.title
			self.url = string.replace(self.url, ' ', '%20')
		else:
			self.url = string.replace(self.url+self.title,' ','%20')
		try:
			url = self.url.encode(self.encode)
		except UnicodeEncodeError:
			url = self.url.encode('utf-8')
		self.progress.set_data(parent_window, _("Searching"), _("Wait a moment"), False)
		retriever = Retriever(url, parent_window, self.progress)
		retriever.start()
		while retriever.isAlive():
			self.progress.pulse()
			if self.progress.status:
				retriever.suspend()
			while gtk.events_pending():
				gtk.main_iteration()
		try:
			ifile = file(retriever.html[0], "rb")
			self.page = ifile.read()
		except IOError:
			pass
		urlcleanup()

class Retriever(threading.Thread):
	def __init__(self, URL, parent_window, progress, destination=None):
		self.URL = URL
		self.html = None
		self.destination = destination
		self.parent_window = parent_window
		self.progress = progress
		self._stopevent = threading.Event()
		self._sleepperiod = 1.0
		threading.Thread.__init__(self, name="Retriever")
	def run(self):
		try:
			self.html = urlretrieve(self.URL, self.destination, self.hook)
			#self.html = urlretrieve(self.URL.encode('utf-8'), self.destination, self.hook)
			if self.progress.status:
				self.html = []
		except IOError:
			self.progress.dialog.hide()
			gutils.urllib_error(_("Connection error"), self.parent_window)
			self.suspend()
	def hook(self,count, blockSize, totalSize):
		if totalSize ==-1:
			pass
		else:
			try:
				downloaded_percentage = min((count*blockSize*100)/totalSize, 100)
			except:
				downloaded_percentage = 100
			if count != 0:
				downloaded_kbyte = int(count * blockSize/1024.0)
				filesize_kbyte = int(totalSize/1024.0)

#
# use own derived URLopener class because we need to set a correct User-Agent
# string for some web sites. The default is 'Python-urllib/<version>'
# which not all sites accepting. (zelluloid.de for example)
#
_uaurlopener = None
def urlretrieve(url, filename=None, reporthook=None, data=None):
    global _uaurlopener
    if not _uaurlopener:
        _uaurlopener = UAFancyURLopener()
    return _uaurlopener.retrieve(url, filename, reporthook, data)

class UAFancyURLopener(FancyURLopener):
	# use Firefox 3.0 User-Agent string from Windows XP
	version = 'Mozilla/5.0 (Windows; U; Windows NT 6.0) Gecko/2008052906 Firefox/3.0'
	
class Progress:
	def __init__(self, window, title = '', message = ''):
		self.status = False
		self.dialog = gtk.Dialog(title, window, gtk.DIALOG_MODAL, ())
		self.dialog.set_urgency_hint(False)
		self.label = gtk.Label()
		self.label.set_markup(message)
		self.dialog.vbox.pack_start(self.label)
		self.progress = gtk.ProgressBar()
		self.progress.set_pulse_step(0.01)
		self.dialog.vbox.pack_start(self.progress, False, False)
		self.button = gtk.Button(_("Cancel"), gtk.STOCK_CANCEL)
		self.button.connect("clicked", self.callback)
		self.dialog.vbox.pack_start(self.button, False, False)
		self.dialog.show_all()
	def callback(self, widget):
		self.dialog.hide()
		self.status = True
	def pulse(self):
		self.progress.pulse()
		time.sleep(0.01)
	def close(self):
		self.dialog.destroy()
	def hide(self):
		self.dialog.hide()
	def set_data(self, parent_window, title, message, showit):
		#self.dialog.set_parent(parent_window)
		self.dialog.set_title(title)
		self.label.set_markup(message)
		if showit == True:
			self.dialog.show()
# vim: fdm=marker



# -*- coding: utf-8 -*-

__revision__ = '$Id: PluginMovieFDb.py 1040 2008-11-15 21:13:49Z mikej06 $'

# Copyright (c) 2006-2007 Piotr Ozarowski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

import gutils
import movie
import string
import re

plugin_name		= 'FDb'
plugin_description	= 'Internetowa baza filmowa'
plugin_url		= 'fdb.pl'
plugin_language		= _('Polish')
plugin_author		= 'Piotr Ożarowski'
plugin_author_email	= '<ozarow+griffith@gmail.com>'
plugin_version		= '1.8'

class Plugin(movie.Movie):
	TRAILER_PATTERN = re.compile('\s*<a\shref=[\'"](.*?)["\'].*?>\(zobacz\szwiastun\)</a>')
	
	def __init__(self, movie_id):
		from md5 import md5
		self.movie_id = md5(movie_id).hexdigest()
		self.encode   = 'utf-8'
		if string.find(movie_id, 'http://') != -1:
			self.url = str(movie_id)
		else:
			self.url = "http://fdb.pl/%s" % str(movie_id)

	def get_image(self):
		self.image_url = gutils.trim(self.page, 'class="moviePosterTable"', '</td>');
		self.image_url = gutils.trim(self.image_url,' src="',"\"\n")
		if self.image_url.endswith('no_picture.png'):
			self.image_url = ''
		else:
			self.image_url = "http://fdb.pl%s" % self.image_url

	def get_o_title(self):
		self.o_title = gutils.trim(self.page, '<h2>', '</h2>')
		self.o_title = gutils.strip_tags(self.o_title)
		if self.o_title == '':
			self.o_title = self.get_title(True)

	def get_title(self, extra=False):
		data = gutils.trim(self.page,'<title>', '</title>')
		tmp = string.find(data, '(')
		if tmp != -1:
			data = data[:tmp]
		if extra is False:
			self.title = data
		else:
			return data

	def get_director(self):
		self.director = ''
		elements = gutils.trim(self.page,'>Reżyseria</div>','<br />')
		elements = string.split(elements, '<div>')
		if elements[0] != '':
			for element in elements:
				element = gutils.trim(element, '>', '</a')
				if element != '':
					self.director += ', ' + element
			self.director = string.replace(self.director[2:], ', &nbsp;&nbsp;&nbsp;(więcej)', '')

	def get_plot(self):
		self.plot = gutils.trim(self.page,'>Opis filmu:</div>','</div>')
		self.plot = self.TRAILER_PATTERN.sub('', self.plot)

	def get_year(self):
		self.year = gutils.trim(self.page,' class="movieYear">(', ')</span>')

	def get_runtime(self):
		self.runtime = gutils.trim(self.page,">Czas trwania:</span>\n        ",' min')

	def get_genre(self):
		self.genre = gutils.trim(self.page,'>Gatunek:</span>','</div>')
		self.genre = string.replace(self.genre, '   ', '')

	def get_cast(self):
		self.cast = gutils.trim(self.page,'>Obsada:</div>', """<tr>
            <td colspan""")
		if self.cast != '':
			self.cast = self.cast.replace('....',_(' as '))
			self.cast = self.cast.replace('  ', '')
			self.cast = self.cast.replace("\n", '')
			self.cast = self.cast.replace('</tr>', "\n")

	def get_classification(self):
		self.classification = gutils.trim(self.page,">Od lat:</span>\n","\n")

	def get_studio(self):
		self.studio = ''

	def get_o_site(self):
		self.o_site = ''

	def get_site(self):
		self.site = self.url

	def get_trailer(self):
		trailer_url = self.TRAILER_PATTERN.findall(self.page)
		if trailer_url:
			self.trailer = trailer_url[0]

	def get_country(self):
		self.country = gutils.trim(self.page,">Kraj:</span>\n",'</div>')
		self.country = string.replace(self.country, "   ", '')

	def get_rating(self):
		self.rating = gutils.trim(self.page, '>Ocena:</span>','/10</span>')
		self.rating = gutils.after(self.rating, 'bold">')
		if self.rating:
			self.rating = str(float(gutils.clean(self.rating)))

class SearchPlugin(movie.SearchMovie):
	def __init__(self):
		self.encode = 'utf-8'
		self.original_url_search	= 'http://fdb.pl/szukaj.php?t=f&s='
		self.translated_url_search	= 'http://fdb.pl/szukaj.php?t=f&s='

	def search(self,parent_window):
		self.open_search(parent_window)
		tmp = string.find(self.page,'<div>Wyniki wyszukiwania dla')
		if tmp == -1:		# already a movie page
			self.page = ''
		else:			# multiple matches
			self.page = gutils.before(self.page[tmp:],'<div id="mapaSerwisu">');
		return self.page

	def get_searches(self):
		if self.page == '':	# movie page already
			self.number_results = 1
			self.ids.append(self.url)
			self.titles.append(self.title)
		else:			# multiple matches
			elements = string.split(self.page,'<div class="searchItem">')
			if len(elements)>0:
				for element in elements:
					tmpId = gutils.trim(element, '<a href="', '"')
					if tmpId.endswith('dodajNowy.php'):
						continue
					self.ids.append(tmpId)
					element = gutils.strip_tags(
						gutils.trim(element, '">', '</div>'))
					element = element.replace("\n", '')
					element = element.replace('   ', '')
					element = element.replace('aka ', ' aka ')
					element = element.replace(' - Oryginalny', '')
					self.titles.append(element)
			else:
				self.number_results = 0



# -*- coding: iso-8859-15 -*-

__revision__ = '$Id: PluginMovieCulturalia.py 389 2006-07-29 18:43:35Z piotrek $'

# Copyright (c) 2006 Pedro D. Sánchez
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

import gutils
import movie
import string

plugin_name		= 'Culturalia'
plugin_description	= 'Base de Datos de Peliculas'
plugin_url		= 'www.culturalianet.com'
plugin_language		= _('Spanish')
plugin_author		= 'Pedro D. Sánchez'
plugin_author_email	= '<pedrodav@gmail.com>'
plugin_version		= '0.2'

class Plugin(movie.Movie):
	def __init__(self, id):
		self.encode='iso-8859-15'
		self.movie_id = id
		self.url = "http://www.culturalianet.com/art/ver.php?art=%s" % str(self.movie_id)

	def get_image(self):
		tmp = string.find(self.page, "<font class = 'titulo2'>")
		if tmp == -1:
			self.image_url = ''
		else:
			self.image_url = 'http://www.culturalianet.com/art/'+gutils.trim(self.page[tmp:], '<img src = "', '"')

	def get_o_title(self):
		self.o_title = gutils.trim(self.page, "<font class = 'titulo2'>", '</i>')
		self.o_title = gutils.after(self.o_title, '<i>')

	def get_title(self):
		self.title = gutils.trim(self.page, "<font class = 'titulo2'>", '. (')

	def get_director(self):
		self.director = gutils.trim(self.page,"<font class = 'titulo3'>Director:", '<br><br>')

	def get_plot(self):
		self.plot = gutils.trim(self.page, '<b>Sinopsis:</b><br>', '<br><br>')

	def get_year(self):
		self.year = gutils.trim(self.page, "<font class = 'titulo2'>", ')</font>')
		self.year = gutils.after(self.year, '. (')

	def get_runtime(self):
		self.runtime = gutils.trim(self.page, "<font class = 'titulo3'>Duración:</font> ", ' minutos')
		if self.runtime == '':
			self.runtime = gutils.trim(self.page, "<font class = 'titulo3'>Duraci&oacute;n:</font> ", ' minutos')

	def get_genre(self):
		self.genre = gutils.trim(self.page, "<font class = 'titulo3'>Género:</font><br>", '<br><br>')

	def get_cast(self):
		self.cast = ''
		self.cast = gutils.trim(self.page, "<font class = 'titulo3'>Actores:</font><br>", '<br><br>')
		self.cast = string.replace(self.cast, '<br>', "\n")
		self.cast = string.strip(gutils.strip_tags(self.cast))

	def get_classification(self):
		self.classification = gutils.trim(self.page, "<font class = 'titulo3'>Calificación moral:</font> ", '<br>')
		if self.classification == '':
			self.classification = gutils.trim(self.page, "<font class = 'titulo3'>Calificaci&oacute;n moral:</font> ", '<br>')

	def get_studio(self):
		self.studio = ''

	def get_o_site(self):
		self.o_site = ''

	def get_site(self):
		self.site = "http://www.culturalianet.com/art/ver.php?art=%s" % str(self.movie_id)

	def get_trailer(self):
		self.trailer = ''

	def get_country(self):
		self.country = gutils.trim(self.page, "<font class = 'titulo3'>Nacionalidad:</font><br>", '<br>')

	def get_rating(self):
		self.rating = gutils.trim(self.page, "Puntuación: <font class = 'titulo3'>", '</font>')
		if self.rating == '':
			self.rating = gutils.trim(self.page, "Puntuaci&oacute;n: <font class = 'titulo3'>", '</font>')
		if self.rating:
			self.rating = str(float(gutils.clean(self.rating)))

class SearchPlugin(movie.SearchMovie):

	def __init__(self):
		self.original_url_search	= 'http://www.culturalianet.com/bus/resu.php?donde=1&texto='
		self.translated_url_search	= 'http://www.culturalianet.com/bus/resu.php?donde=1&texto='
		self.encode = 'iso-8859-15'

	def search(self,parent_window):
		self.open_search(parent_window)
		self.sub_search()
		return self.page

	def sub_search(self):
		self.page = gutils.trim(self.page, '<b>ARTICULOS ENCONTRADOS:</b>', '</table>')
		self.page = gutils.after(self.page, '</td></tr><tr><td><b>')
		#self.page = self.page.decode('iso-8859-15')

	def get_searches(self):
		elements = string.split(self.page, '<td><b>')

		if (elements[0]<>''):
			for element in elements:
				self.ids.append(gutils.trim(element, 'ver.php?art=',"'"))
				self.titles.append(gutils.strip_tags(gutils.convert_entities(gutils.trim(element, "target='_top'>", '</a>'))))



# -*- coding: UTF-8 -*-

__revision__ = '$Id: PluginMovieIMDB.py 176 2006-02-01 12:07:26Z iznogoud $'

# Copyright (c) 2005-2006 Vasco Nunes, Piotr Ozarowski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, orprint
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

import gutils
import movie
import string

plugin_name = "Allocine"
plugin_description = "Internet Movie Database"
plugin_url = "www.allocine.fr"
plugin_language = _("French")
plugin_author = "Pierre-Luc Levy"
plugin_author_email = ""
plugin_version = "0.6"

class Plugin(movie.Movie):
	def __init__(self, id):
		self.movie_id = id
		self.url = "http://www.allocine.fr/film/fichefilm_gen_cfilm=%s.html" % str(self.movie_id)

	def get_image(self):
		self.image_url = gutils.trim(self.page,"Poster","Date de sor")
		self.image_url = gutils.after(self.image_url,"activerlientexte.inc")
		self.image_url = gutils.trim(self.image_url,"<img src=\"","\"")

	def get_o_title(self):
		self.o_title = ""
		self.o_title = gutils.trim(self.page,"Titre original : <i>","</i>")
		if (self.o_title==''):
			self.o_title = gutils.trim(self.page,"<title>","</title>")

	def get_title(self):
		self.title = gutils.trim(self.page,"<title>","</title>")

	def get_director(self):
		self.director = gutils.trim(self.page,"<h4>Réalisé par ","</a></h4>")

	def get_plot(self):
		self.plot = gutils.trim(self.page,"Synopsis</b></h3></td></tr></table>","</h4>")
		self.plot = gutils.after(self.plot,"<h4>")

	def get_year(self):
		self.year = gutils.trim(self.page,"Année de production : ","</h4>")

	def get_runtime(self):
		self.runtime = ""
		self.runtime = gutils.trim(self.page,"<h4>Durée : ","min.</h4>&nbsp;")
		if self.runtime:
			self.runtime = str (int(gutils.before(self.runtime,"h"))*60 + int(gutils.after(self.runtime,"h")))

	def get_genre(self):
		self.genre = gutils.trim(self.page,"<h4>Genre : ","</h4>")
		self.genre = gutils.strip_tags(self.genre)

	def get_cast(self):
		self.cast = ""
		self.cast = gutils.trim(self.page,"<h4>Avec ","</h4>")
		self.cast = gutils.strip_tags(self.cast)
		self.cast = string.replace(self.cast,", ", "\n")

	def get_image(self):
		self.classification = ""

	def get_studio(self):
		self.studio = ""

	def get_o_site(self):
		self.o_site = ""

	def get_site(self):
		self.site = "http://www.allocine.fr/film/fichefilm_gen_cfilm=%s.html" % self.movie_id

	def get_trailer(self):
		self.trailer = "http://www.allocine.fr/film/video_gen_cfilm=%s.html" % self.movie_id

	def get_country(self):
		self.country = gutils.trim(self.page,"<h4>Film ",".</h4>&nbsp;")

	def get_rating(self):
		self.rating = gutils.trim(self.page, "Spectateurs</a> ", "</h4>")
		self.rating = gutils.trim(self.rating, "etoile_", ".gif")
		if self.rating:
			self.rating = str(float(int(self.rating)*2.25))

class SearchPlugin(movie.SearchMovie):

	def __init__(self):
		self.original_url_search	= "http://www.allocine.fr/recherche/?motcle="
		self.translated_url_search	= "http://www.allocine.fr/recherche/?motcle="

	def search(self,parent_window):
		self.open_search(parent_window)
		self.sub_search()
		return self.page

	def sub_search(self):
		self.page = gutils.trim(self.page,"Recherche : <b>", "<h3><b>Articles <h4>");

	def get_searches(self):
		elements = string.split(self.page,"<td colspan=\"2\" height=\"1\" valign=\"top\"><hr /></td>")
		if (elements[0]<>''):
			for element in elements:
				self.ids.append(gutils.trim(element,"/film/fichefilm_gen_cfilm=",".html"))
				self.titles.append(gutils.strip_tags(gutils.convert_entities(gutils.trim(element,"link1\">","</a>"))))



# -*- coding: UTF-8 -*-

__revision__ = '$Id: dbupgrade.py 1040 2008-11-15 21:13:49Z mikej06 $'

# Copyright (c) 2005-2008 Vasco Nunes, Piotr Ożarowski
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

# You may use and distribute this software under the terms of the
# GNU General Public License, version 2 or later

from sqlalchemy import *
import os.path
import gutils
import gtk


def upgrade_database(self, version):
	"""Create new db or update existing one to current format"""
	if version == 0:
		self.Configuration.mapper.mapped_table.create()
		self.Configuration.mapper.mapped_table.insert().execute(param='version', value=self.version)
		self.Volume.mapper.mapped_table.create()
		self.Collection.mapper.mapped_table.create()
		self.Medium.mapper.mapped_table.create()
		self.Medium.mapper.mapped_table.insert().execute(name='DVD')
		self.Medium.mapper.mapped_table.insert().execute(name='DVD-R')
		self.Medium.mapper.mapped_table.insert().execute(name='DVD-RW')
		self.Medium.mapper.mapped_table.insert().execute(name='DVD+R')
		self.Medium.mapper.mapped_table.insert().execute(name='DVD+RW')
		self.Medium.mapper.mapped_table.insert().execute(name='DVD-RAM')
		self.Medium.mapper.mapped_table.insert().execute(name='CD')
		self.Medium.mapper.mapped_table.insert().execute(name='CD-RW')
		self.Medium.mapper.mapped_table.insert().execute(name='VCD')
		self.Medium.mapper.mapped_table.insert().execute(name='SVCD')
		self.Medium.mapper.mapped_table.insert().execute(name='VHS')
		self.Medium.mapper.mapped_table.insert().execute(name='BETACAM')
		self.Medium.mapper.mapped_table.insert().execute(name='LaserDisc')
		self.ACodec.mapper.mapped_table.create()
		self.ACodec.mapper.mapped_table.insert().execute(name='AC-3 Dolby audio')
		self.ACodec.mapper.mapped_table.insert().execute(name='OGG')
		self.ACodec.mapper.mapped_table.insert().execute(name='MP3')
		self.ACodec.mapper.mapped_table.insert().execute(name='MPEG-1')
		self.ACodec.mapper.mapped_table.insert().execute(name='MPEG-2')
		self.ACodec.mapper.mapped_table.insert().execute(name='AAC')
		self.ACodec.mapper.mapped_table.insert().execute(name='Windows Media Audio')
		self.VCodec.mapper.mapped_table.create()
		self.VCodec.mapper.mapped_table.insert().execute(name='MPEG-1')
		self.VCodec.mapper.mapped_table.insert().execute(name='MPEG-2')
		self.VCodec.mapper.mapped_table.insert().execute(name='XviD')
		self.VCodec.mapper.mapped_table.insert().execute(name='DivX')
		self.VCodec.mapper.mapped_table.insert().execute(name='H.264')
		self.VCodec.mapper.mapped_table.insert().execute(name='RealVideo')
		self.VCodec.mapper.mapped_table.insert().execute(name='QuickTime')
		self.VCodec.mapper.mapped_table.insert().execute(name='Windows Media Video')
		self.AChannel.mapper.mapped_table.create()
		self.AChannel.mapper.mapped_table.insert().execute(name='mono')
		self.AChannel.mapper.mapped_table.insert().execute(name='stereo')
		self.AChannel.mapper.mapped_table.insert().execute(name='5.1')
		self.AChannel.mapper.mapped_table.insert().execute(name='7.1')
		self.SubFormat.mapper.mapped_table.create()
		self.SubFormat.mapper.mapped_table.insert().execute(name='DVD VOB')
		self.SubFormat.mapper.mapped_table.insert().execute(name='MPL2 (.txt)')
		self.SubFormat.mapper.mapped_table.insert().execute(name='MicroDVD (.sub)')
		self.SubFormat.mapper.mapped_table.insert().execute(name='SubRip (.srt)')
		self.SubFormat.mapper.mapped_table.insert().execute(name='SubViewer2 (.sub)')
		self.SubFormat.mapper.mapped_table.insert().execute(name='Sub Station Alpha (.ssa)')
		self.SubFormat.mapper.mapped_table.insert().execute(name='Advanced Sub Station Alpha (.ssa)')
		self.Person.mapper.mapped_table.create()
		self.Movie.mapper.mapped_table.create()
		self.Loan.mapper.mapped_table.create()
		self.Lang.mapper.mapped_table.create()
		self.Lang.mapper.mapped_table.insert().execute(name=_('Brazilian Portuguese'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Bulgarian'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Catalan'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Czech'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Danish'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Dutch'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('English'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Estonian'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('French'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('German'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Greek'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Hungarian'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Indonesian'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Italian'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Japanese'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Korean'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Norwegian Bokmal'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Occitan'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Pashto'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Polish'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Portuguese'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Russian'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Simplified Chinese'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Slovak'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Spanish'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Swedish'))
		self.Lang.mapper.mapped_table.insert().execute(name=_('Turkish'))
		self.MovieLang.mapper.mapped_table.create()
		self.Tag.mapper.mapped_table.create()
		self.Tag.mapper.mapped_table.insert().execute(name=_('Favourite'))
		self.MovieTag.mapper.mapped_table.create()
		#self.metadata.commit()
		return True # upgrade process finished
	if version == 1: # fix changes between v1 and v2
		version+=1
		self.metadata.engine.execute("UPDATE loans SET return_date = '2007-01-01' WHERE return_date='None';")
		db_version = self.Configuration.get_by(param='version')
		db_version.value = version
		db_version.update()
		db_version.flush()
	#if version == 2:	# fix changes between v2 and v3
	#	version+=1
	#	self.Configuration.get_by(param='version').value = version
	return True


# ---------------------------------------------------
# for Griffith <= 0.6.2 compatibility
# ---------------------------------------------------

def convert_from_old_db(self, source_file, destination_file):	#{{{
	print 'Converting old database - it can take several minutes...'
	gutils.info(self,_("Griffith will now convert your database to the new format. This can take several minutes if you have a large database."))
	from sqlalchemy.orm import clear_mappers
	from sql import GriffithSQL
	from gutils import digits_only
	import os

	if not os.path.isfile(source_file):
		return False
	if open(source_file).readline()[:47] == '** This file contains an SQLite 2.1 database **':
		try:
			import sqlite
			from sqlite import DatabaseError
		except ImportError:
			print 'Old DB conversion: please install pysqlite legacy (v1.0)'
			gutils.warning(self,_("Old DB conversion: please install pysqlite legacy (v1.0)"))
			return False
	else:
		try:	# Python 2.5
			from sqlite3 import dbapi2 as sqlite
			from sqlite3.dbapi2 import DatabaseError
		except ImportError: # Python < 2.5 - try to use pysqlite2
			from pysqlite2 import dbapi2 as sqlite
			from pysqlite2.dbapi2 import DatabaseError

	if os.path.isfile(destination_file):
		# rename destination_file if it already exist
		i = 1
		while True:
			if os.path.isfile("%s_%s" % (destination_file, i)):
				i += 1
			else:
				break
		os.rename(destination_file, "%s_%s" % (destination_file, i))

	try:
		old_db = sqlite.connect(source_file)
	except DatabaseError, e:
		if str(e) == 'file is encrypted or is not a database':
			print 'Your database is most probably in wrong SQLite format, please convert it to SQLite3:'
			print '$ sqlite ~/.griffith/griffith.gri .dump | sqlite3 ~/.griffith/griffith.gri3'
			print '$ mv ~/.griffith/griffith.gri{,2}'
			print '$ mv ~/.griffith/griffith.gri{3,}'
			print 'or install pysqlite in version 1.0'
			gutils.warning(self,_("Your database is most probably in SQLite2 format, please convert it to SQLite3"))
		else:
			raise
		return False

	old_cursor = old_db.cursor()

	# fix old database
	old_cursor.execute("UPDATE movies SET media = '1' WHERE media = 'DVD';")
	old_cursor.execute("UPDATE movies SET media = '2' WHERE media = 'DVD-R';")
	old_cursor.execute("UPDATE movies SET media = '3' WHERE media = 'DVD-RW';")
	old_cursor.execute("UPDATE movies SET media = '4' WHERE media = 'DVD+R';")
	old_cursor.execute("UPDATE movies SET media = '5' WHERE media = 'DVD+RW';")
	old_cursor.execute("UPDATE movies SET media = '6' WHERE media = 'DVD-RAM';")
	old_cursor.execute("UPDATE movies SET media = '7' WHERE media = 'DivX';")
	old_cursor.execute("UPDATE movies SET media = '7' WHERE media = 'DIVX';")
	old_cursor.execute("UPDATE movies SET media = '7' WHERE media = 'XviD';")
	old_cursor.execute("UPDATE movies SET media = '7' WHERE media = 'XVID';")
	old_cursor.execute("UPDATE movies SET media = '7' WHERE media = 'WMV';")
	old_cursor.execute("UPDATE movies SET media = '9' WHERE media = 'VCD';")
	old_cursor.execute("UPDATE movies SET media = '10' WHERE media = 'SVCD'; 	")
	old_cursor.execute("UPDATE movies SET media = '11' WHERE media = 'VHS';")
	old_cursor.execute("UPDATE movies SET media = '12' WHERE media = 'BETACAM';")
	old_cursor.execute("UPDATE movies SET collection_id=0 WHERE collection_id<1")
	old_cursor.execute("UPDATE movies SET volume_id=0 WHERE volume_id<1")
	old_cursor.execute("UPDATE movies SET color=NULL WHERE color<1 OR color='' OR color>3")
	old_cursor.execute("UPDATE movies SET condition=NULL WHERE condition<0 OR condition='' OR condition>5")
	old_cursor.execute("UPDATE movies SET layers=NULL WHERE layers<0 OR layers='' OR layers>4")
	old_cursor.execute("UPDATE movies SET region=NULL WHERE region='' OR region=2 OR region<0 OR region>8")
	old_cursor.execute("UPDATE movies SET year=NULL WHERE year<1900 or year>2007")
	old_cursor.execute("UPDATE movies SET rating = 0 WHERE rating NOT IN (0,1,2,3,4,5,6,7,8,9,10);") # rating>10 doesn't work with some DB
	old_cursor.execute("UPDATE movies SET runtime = NULL WHERE runtime > 10000;") # remove strings
	old_cursor.execute("UPDATE loans SET return_date=NULL WHERE return_date=''")
	old_cursor.execute("UPDATE loans SET return_date=NULL WHERE return_date='None'")
	old_cursor.execute("DELETE FROM loans WHERE date='' OR date ISNULL")
	old_cursor.execute("DELETE FROM volumes WHERE name = ''")
	old_cursor.execute("DELETE FROM volumes WHERE name = 'None'")
	old_cursor.execute("DELETE FROM collections WHERE name = ''")
	old_cursor.execute("DELETE FROM collections WHERE name = 'None'")
	old_cursor.execute("DELETE FROM languages WHERE name = ''")
	
	self.config.set('type','sqlite', section='database')
	self.config.set('file', 'griffith.db', section='database')
	self.config['posters'] = 'posters'
	self.config.set('color', 0, section='defaults')
	self.config.set('condition', 0, section='defaults')
	self.config.set('layers', 0, section='defaults')
	self.config.set('media', 0, section='defaults')
	self.config.set('region', 0, section='defaults')
	self.config.set('vcodec', 0, section='defaults')
	self.locations['posters'] = os.path.join(self.locations['home'], 'posters')
	new_db = GriffithSQL(self.config, self.debug, self.locations['home'])

	# collections
	collection_mapper = {'':None, u'':None, 0:None, '0':None, -1:None, '-1':None}
	old_cursor.execute("SELECT id, name FROM collections;") # loaned status will be set later - buggy databases :-(
	for i in old_cursor.fetchall():
		o = new_db.Collection(name=i[1])
		try:
			o.save(); o.flush()
		except Exception, e:
			self.debug.show(str(e))
			continue
		collection_mapper[i[0]] = o.collection_id
	
	# volumes
	volume_mapper = {'':None, u'':None, 0:None, '0':None, -1:None, '-1':None}
	old_cursor.execute("SELECT id, name FROM volumes;") # loaned status will be set later - buggy databases :-(
	for i in old_cursor.fetchall():
		o = new_db.Volume(name=i[1])
		try:
			o.save(); o.flush()
		except Exception, e:
			self.debug.show(str(e))
			continue
		volume_mapper[i[0]] = o.volume_id

	# people
	person_mapper = {}
	old_cursor.execute("SELECT id, name, email, phone FROM people;")
	for i in old_cursor.fetchall():
		o = new_db.Person(name=i[1], email=i[2], phone=i[3])
		try:
			o.save(); o.flush()
		except Exception, e:
			self.debug.show(str(e))
			continue
		person_mapper[i[0]] = o.person_id
	
	# languages
	language_mapper = {'':None, u'':None, 0:None, '0':None, -1:None, '-1':None}
	old_cursor.execute("SELECT id, name FROM languages;")
	for i in old_cursor.fetchall():
		o = new_db.Lang.get_by(name=i[1])
		if o is not None:
			language_mapper[i[0]] = o.lang_id
		else:
			o = new_db.Lang(name=i[1])
			try:
				o.save(); o.flush()
			except Exception, e:
				self.debug.show(str(e))
				continue
			language_mapper[i[0]] = o.lang_id

	# media
	medium_mapper = {'':None, u'':None, 0:None, '0':None, -1:None, '-1':None}
	old_cursor.execute("SELECT id, name FROM media;")
	for i in old_cursor.fetchall():
		o = new_db.Medium.get_by(name=i[1])
		if o is not None:
			medium_mapper[i[0]] = o.medium_id
		else:
			o = new_db.Medium(name=i[1])
			try:
				o.save(); o.flush()
			except Exception, e:
				self.debug.show(str(e))
				continue
			medium_mapper[i[0]] = o.medium_id
	
	# tags
	tag_mapper = {}
	old_cursor.execute("SELECT id, name FROM tags;")
	for i in old_cursor.fetchall():
		o = new_db.Tag.get_by(name=i[1])
		if o is not None:
			tag_mapper[i[0]] = o.tag_id
		else:
			o = new_db.Tag(name=i[1])
			try:
				o.save(); o.flush()
			except Exception, e:
				self.debug.show(str(e))
				continue
			tag_mapper[i[0]] = o.tag_id
	
	# movies
	movie_mapper = {}
	old_cursor.execute("""
		SELECT id, volume_id, collection_id, original_title, title, director,
			number, image, plot, country, year, runtime, classification,
			genre, studio, site, imdb, actors, trailer, rating, loaned,
			media, num_media, obs, seen, region, condition, color, layers
		FROM movies ORDER BY number;""")
	for i in old_cursor.fetchall():
		o = new_db.Movie()
		o.number = digits_only(i[6])
		if volume_mapper.has_key(i[1]):
			o.volume_id = volume_mapper[i[1]]
		if collection_mapper.has_key(i[2]):
			o.collection_id = collection_mapper[i[2]]
		o.o_title = i[3][:255]
		o.title = i[4][:255]
		o.director = i[5][:255]
		o.image = i[7][:128]
		o.plot = i[8]
		o.country = i[9][:128]
		o.year = digits_only(i[10])
		o.runtime = digits_only(i[11])
		o.classification = i[12][:128]
		o.genre = i[13][:128]
		o.studio = i[14][:128]
		o.o_site = i[15][:255]
		o.site = i[16][:255]
		o.cast = i[17]
		o.trailer = i[18][:255]
		o.rating = digits_only(i[19])
		#o.loaned = bool(i[20]) # updated later
		if medium_mapper.has_key(int(i[21])):
			o.medium_id = medium_mapper[int(i[21])]
		o.media_num = digits_only(i[22])
		o.notes = i[23]
		o.seen = bool(i[24])
		o.region = digits_only(i[25])
		o.cond = digits_only(i[26], 5)
		o.color = digits_only(i[27], 3)
		o.layers = digits_only(i[28], 4)
		
		try:
			o.save(); o.flush()
		except Exception, e:
			self.debug.show(str(e))
			continue
		movie_mapper[i[0]] = o.movie_id

	# movie tag
	old_cursor.execute("SELECT movie_id, tag_id FROM movie_tag WHERE movie_id IN (SELECT id FROM movies);")
	for i in old_cursor.fetchall():
		o = new_db.MovieTag.get_by(movie_id=movie_mapper[i[0]], tag_id=tag_mapper[i[1]])
		if o is None:
			m = new_db.Movie.get_by(movie_id=movie_mapper[i[0]])
			t = new_db.Tag.get_by(tag_id=tag_mapper[i[1]])
			t.save()
			m.tags.append(t)
			try:
				m.save(); m.flush()
			except Exception, e:
				self.debug.show(str(e))
				continue
	
	# movie lang
	old_cursor.execute("SELECT movie_id, lang_id, type FROM movie_lang WHERE movie_id IN (SELECT id FROM movies);")
	for i in old_cursor.fetchall():
		o = new_db.MovieLang.get_by(movie_id=movie_mapper[i[0]], lang_id=language_mapper[i[1]], type=i[2])
		if o is None:
			m = new_db.Movie.get_by(movie_id=movie_mapper[i[0]])
			l = new_db.MovieLang(lang_id=language_mapper[i[1]], type=i[2])
			l.save()
			m.languages.append(l)
			try:
				m.save(); m.flush()
			except Exception, e:
				self.debug.show(str(e))
				continue

	# loans
	old_cursor.execute("SELECT person_id, movie_id, volume_id, collection_id, date, return_date FROM loans;")
	for i in old_cursor.fetchall():
		vol = col = None
		not_returned = i[5] is None

		if int(i[2]) > 0:
			try:
				vol = new_db.Volume.get_by(volume_id=volume_mapper[i[2]])
			except Exception, e:
				self.debug.show(str(e))
				continue
		if int(i[3]) > 0:
			try:
				col = new_db.Collection.get_by(collection_id=collection_mapper[i[3]])
			except Exception, e:
				self.debug.show(str(e))
				continue
		if int(i[1]) == 0:
			if vol is not None and len(vol.movies)>0:
				m = vol.movies[0]
			elif col is not None and len(col.movies)>0:
				m = col.movies[0]
			else:
				self.debug.show("Cannot find associated movie for this loan (%s)" % i)
				continue
		else:
			try:
				m = new_db.Movie.get_by(movie_id=movie_mapper[i[1]])
			except Exception, e:
				self.debug.show(str(e))
				continue
		
		l = new_db.Loan()
		l.person_id = person_mapper[i[0]]
		l.date = str(i[4])[:10]
		if not_returned:
			m.loaned = True
			l.return_date = None
		else:
			l.return_date = str(i[5])[:10]
		
		# update volume / collection status
		if int(i[2]) > 0:
			l.volume_id = volume_mapper[i[2]]
			if not_returned:
				vol.loaned = True
				vol.save()
		if int(i[3]) > 0:
			l.collection_id = collection_mapper[i[3]]
			if not_returned:
				col.loaned = True
				col.save()
		l.save();
		m.loans.append(l)
		try:
			m.flush()
		except Exception, e:
			self.debug.show(str(e))
			continue
	clear_mappers()
	return True
#}}}
# vim: fdm=marker
